/* 
 * File:   GinkoSolver.cpp
 * Author: Martin Kjellqvist <martin.kjellqvist@miun.se>
 * 
 * Created on March 19, 2018, 1:19 PM
 */

#include "GinkoSolver.h"
#include "FixedAnimationContainer.h"
#include <queue>
#include <stdexcept>
#include <iostream>

struct tile_helper {
    bool wide;
    bool high;
};


static solution_t best_moves;

/**
 * Find the available moves to an open square.
 * 
 * Will not include moves where open_row, open_col is the second open square.
 * i.e. bottom open to square/vertical (left/right moves), 
 * right open to square/horizontal (up/down moves)
 * 
 * @param bh
 * @param open_row
 * @param open_col
 * @return 
 */
std::vector<Move> moves_to_open(const board_helper& bh, int open_row, int open_col) {
    std::vector<Move> result;
    // Squares are easy, look up
    if (open_row >= 1 && bh.board[open_row - 1][open_col] == SQUARE) {
        Move m;
        m.from_row_col = {open_row - 1, open_col};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // Squares + vertical are easy, look down
    if (open_row < 4 && (
            bh.board[open_row + 1][open_col] == SQUARE ||
            bh.board[open_row + 1][open_col] == VERTICAL)) {
        Move m;
        m.from_row_col = {open_row + 1, open_col};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // Squares  are easy, look left 
    if (open_col >= 1 && bh.board[open_row][open_col - 1] == SQUARE) {
        Move m;
        m.from_row_col = {open_row, open_col - 1};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // Squares + horizontal are easy, look right
    if (open_col < 3 && (
            bh.board[open_row][open_col + 1] == SQUARE ||
            bh.board[open_row][open_col + 1] == HORIZONTAL)) {
        Move m;
        m.from_row_col = {open_row, open_col + 1};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // bigger stuff. Move piece two spots away.
    // look up for vertical
    if (open_row >= 2 && bh.board[open_row - 2][open_col] == VERTICAL) {
        Move m;
        m.from_row_col = {open_row - 2, open_col};
        m.to_row_col = {open_row - 1, open_col};
        result.push_back(m);
    }
    // bigger stuff. Move piece two spots away.
    // look left for horizontal    
    if (open_col >= 2 && bh.board[open_row][open_col - 2] == HORIZONTAL) {
        Move m;
        m.from_row_col = {open_row, open_col - 2};
        m.to_row_col = {open_row, open_col - 1};
        result.push_back(m);
    }
    // double_check open spots.
    // look up for horizontal look right for empty.
    // if horizontal is above, col is assumed <= 2 (no invalid board)
    if (open_row >= 1 && bh.board[open_row - 1][open_col] == HORIZONTAL
            && bh.board[open_row][open_col + 1] == EMPTY) {
        Move m;
        m.from_row_col = {open_row - 1, open_col};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // double_check open spots.
    // look down for horizontal look right for empty.
    // if horizontal is below, col is assumed <= 2 (no invalid board)
    if (open_row <= 3 && bh.board[open_row + 1][open_col] == HORIZONTAL
            && bh.board[open_row][open_col + 1] == EMPTY) {
        Move m;
        m.from_row_col = {open_row + 1, open_col};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // double_check open spots.
    // look left for vertical look below for empty.
    // if vertical is left, row is assumed <= 3 (no invalid board)
    if (open_col >= 1 && bh.board[open_row][open_col - 1] == VERTICAL
            && bh.board[open_row + 1][open_col] == EMPTY) {
        Move m;
        m.from_row_col = {open_row, open_col - 1};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // double_check open spots.
    // look right for vertical look down for empty.
    // if vertical is to the right, row is assumed <= 3 (no invalid board)
    if (open_col <= 2 && bh.board[open_row][open_col + 1] == VERTICAL
            && bh.board[open_row + 1][open_col] == EMPTY) {
        Move m;
        m.from_row_col = {open_row, open_col + 1};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }

    // Big stuff
    // Look up
    if (open_row >= 2 && bh.board[open_row - 2][open_col] == BIG
            && bh.board[open_row][open_col + 1] == EMPTY) {
        Move m;
        m.from_row_col = {open_row - 2, open_col};
        m.to_row_col = {open_row - 1, open_col};
        result.push_back(m);
    }
    // Big stuff
    // Look down
    if (open_row <= 2 && bh.board[open_row + 1][open_col] == BIG
            && bh.board[open_row][open_col + 1] == EMPTY) {
        Move m;
        m.from_row_col = {open_row + 1, open_col};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    // Big stuff
    // Look left
    if (open_col >= 2 && bh.board[open_row][open_col - 2] == BIG
            && bh.board[open_row + 1][open_col] == EMPTY) {
        Move m;
        m.from_row_col = {open_row, open_col - 2};
        m.to_row_col = {open_row, open_col - 1};
        result.push_back(m);
    }
    // Big stuff
    // Look right
    if (open_col <= 1 && bh.board[open_row][open_col + 1] == BIG
            && bh.board[open_row + 1][open_col] == EMPTY) {
        Move m;
        m.from_row_col = {open_row, open_col + 1};
        m.to_row_col = {open_row, open_col};
        result.push_back(m);
    }
    return result;
}

std::vector<Move> find_moves(const board_helper& bh) {
    std::vector<Move> result;

    for (int row = 0; row < 5; row++) {
        for (int col = 0; col < 4; col++) {
            if (bh.board[row][col] == EMPTY) {
                auto moves = moves_to_open(bh, row, col);
                result.insert(result.end(), moves.begin(), moves.end());
            }
        }
    }
    return result;
}

/**
 * Invalid moves creates invalid states.
 * 
 * Modifies board in-place.
 * 
 * @param bh
 * @param m
 * @return 
 */
void move(board_helper& bh, Move m) {
    // Squares are easy, just swap
    if (bh.board[m.from_row_col.first][m.from_row_col.second] == EMPTY) {
        std::cerr << "Moving from open\n";
    }
    if (bh.board[m.from_row_col.first][m.from_row_col.second] == SQUARE) {
        std::swap(bh.board[m.from_row_col.first][m.from_row_col.second],
                bh.board[m.to_row_col.first][m.to_row_col.second]);
        return;
    }
    // Verticals are easy, just swap and preserve empty.
    if (bh.board[m.from_row_col.first][m.from_row_col.second] == VERTICAL) {
        // left_right, cols are different, swap + swap below (none + empty)
        // + up. Same swap
        if (m.from_row_col.second != m.to_row_col.second ||
                m.from_row_col.first > m.to_row_col.first) {
            std::swap(bh.board[m.from_row_col.first][m.from_row_col.second],
                    bh.board[m.to_row_col.first][m.to_row_col.second]);
            std::swap(bh.board[m.from_row_col.first + 1][m.from_row_col.second],
                    bh.board[m.to_row_col.first + 1][m.to_row_col.second]);
            return;
        }
        // down, from = empty, to = vertical, to row +1 = none
        bh.board[m.from_row_col.first][m.from_row_col.second] = EMPTY;
        bh.board[m.to_row_col.first][m.to_row_col.second] = VERTICAL;
        bh.board[m.to_row_col.first + 1][m.to_row_col.second] = NONE;
        return;
    }
    // Horizontals are easy, just swap and preserve empty.
    if (bh.board[m.from_row_col.first][m.from_row_col.second] == HORIZONTAL) {
        // up_down, rows are different, swap + swap right(none + empty)
        // + left. Same swap
        if (m.from_row_col.first != m.to_row_col.first ||
                m.from_row_col.second > m.to_row_col.second) {
            std::swap(bh.board[m.from_row_col.first][m.from_row_col.second],
                    bh.board[m.to_row_col.first][m.to_row_col.second]);
            std::swap(bh.board[m.from_row_col.first][m.from_row_col.second + 1],
                    bh.board[m.to_row_col.first][m.to_row_col.second + 1]);
            return;
        }
        // right, from = empty, to = horizontal, to col + 1 = none
        bh.board[m.from_row_col.first][m.from_row_col.second] = EMPTY;
        bh.board[m.to_row_col.first][m.to_row_col.second] = HORIZONTAL;
        bh.board[m.to_row_col.first][m.to_row_col.second + 1] = NONE;
        return;
    }
    // Bigs are complex
    if (bh.board[m.from_row_col.first][m.from_row_col.second] == BIG) {
        // set BIG
        bh.board[m.to_row_col.first][m.to_row_col.second] = BIG;
        bh.board[m.to_row_col.first][m.to_row_col.second + 1] = NONE;
        bh.board[m.to_row_col.first + 1][m.to_row_col.second] = NONE;
        bh.board[m.to_row_col.first + 1][m.to_row_col.second + 1] = NONE;

        // up, 
        if (m.from_row_col.first > m.to_row_col.first) {
            bh.board[m.to_row_col.first + 2][m.to_row_col.second] = EMPTY;
            bh.board[m.to_row_col.first + 2][m.to_row_col.second + 1] = EMPTY;
            return;
        }
        // down, 
        if (m.from_row_col.first < m.to_row_col.first) {
            bh.board[m.from_row_col.first][m.to_row_col.second] = EMPTY;
            bh.board[m.from_row_col.first][m.to_row_col.second + 1] = EMPTY;
            return;
        }
        // left, make sure the empties are ok.
        if (m.from_row_col.second > m.to_row_col.second) {
            bh.board[m.to_row_col.first][m.to_row_col.second + 2] = EMPTY;
            bh.board[m.to_row_col.first + 1][m.to_row_col.second + 2] = EMPTY;
            return;
        }
        // right, make sure the empties are ok.
        bh.board[m.from_row_col.first][m.from_row_col.second] = EMPTY;
        bh.board[m.from_row_col.first + 1][m.from_row_col.second] = EMPTY;
        return;
    }
    std::cerr << "Silly move\n";
    throw std::logic_error("Silly move detected");
}

board_state_t state_for_board(const board_helper& bh) {
    std::string result_string;
    for (int row = 0; row < 5; row++) {
        for (int col = 0; col < 4; col++) {
            result_string += state_for_tile(bh.board[row][col]);
        }
    }
    return result_string;
}

board_helper board_for_state(board_state_t state) {
    auto tiles = tiles_for_state(state);
    board_helper result;
    auto tiles_it = tiles.begin();
    for (int row = 0; row < 5; row++) {
        for (int col = 0; col < 4; col++) {
            // default to none.
            result.board[row][col] = NONE;
            auto tile = *tiles_it;
            if (row > 0) {
                // look up to see if hindrances are there
                if (result.board[row - 1][col] == VERTICAL) {
                    continue;
                }
                if (result.board[row - 1][col] == BIG) {
                    result.board[row][++col] = NONE;
                    continue;
                }
            }
            // nope, move on
            result.board[row][col] = *tiles_it++;
            if (tile == BIG || tile == HORIZONTAL) {
                result.board[row][++col] = NONE;
            }
        }
    }
    return result;
}

bool is_solved(const board_helper& b) {
    return b.board[3][1] == BIG;
}

struct move_state_depth {
    Move move;
    int depth;
    board_state_t state;
};

void print_move(const Move& m) {
    std::cout << "Row Col From: [" <<
            m.from_row_col.first + 0 << ", " <<
            m.from_row_col.second + 0 << "] To [" <<
            m.to_row_col.first + 0 << ", " <<
            m.to_row_col.second + 0 << "]\n";
}

void print_board(const board_helper& bh) {
    auto tile_type_char = [](TileType tt) {
        switch (tt) {
            case SQUARE:
                return "S";
            case VERTICAL:
                return "V";
            case HORIZONTAL:
                return "H";
            case BIG:
                return "B";
            case EMPTY:
                return "E";
            default:
                return "_";
        }
    };
    for (int row = 0; row < 5; row++) {
        for (int col = 0; col < 4; col++) {
            std::cout << tile_type_char(bh.board[row][col]);
        }
        std::cout << "\n";
    }
}

void solve(board_state_t state) {
    // create board from state
    // if in best_moves, done
    // check if success
    // find possible moves,
    // add to queue with depth if not visited.
    // BFS style
    // repeat with popped queue state.
    using move_depth = std::pair<Move, int>;
    using moves = std::vector<move_depth>;

    auto starting_state = state;

    std::map<board_state_t, moves> visited;

    auto put_to_visited = [&visited](board_state_t state, Move m, int depth) {
        if (visited.count(state)) {
            move_depth md = {m, depth};
            visited[state].push_back(md);
        } else {
            moves init;
            move_depth md = {m, depth};
            init.push_back(md);
            visited.insert({state, init});
        }
    };

    std::queue<move_state_depth> move_queue;

    board_helper bh = board_for_state(state);

    auto available_moves = find_moves(bh);
    for (auto move : available_moves) {
        int depth = 0;
        move_queue.push({move, depth, state});
        put_to_visited(state, move, depth);
    }

    auto current_move = move_queue.front();
    while (!move_queue.empty()) {
        current_move = move_queue.front();
        move_queue.pop();
        board_helper current_board = board_for_state(current_move.state);

        move(current_board, current_move.move);
        //print_board(new_board);

        auto state_after_move = state_for_board(current_board);
        if (visited.count(state_after_move)) {
            continue;
        }
        if (is_solved(current_board)) {
            put_to_visited(state_after_move, current_move.move, current_move.depth);
            state = state_after_move;
            break;
        }

        if (best_moves.count(state_after_move)) {
            return;
        }

        //std::cout << "Adding: " << new_state << "\n";

        available_moves = find_moves(current_board);
        for (auto new_move : available_moves) {
            move_queue.push({new_move, current_move.depth + 1, state_after_move});
            put_to_visited(state_after_move, new_move, current_move.depth + 1);
        }

    }

    // depth is local for this particular iteration of solver.
    // optimum path is global/best_moves
    if (move_queue.empty()) {
        std::cout << "Empty move queue\n";
    }

    auto reverse_move = [](Move & m) {
        std::swap(m.from_row_col, m.to_row_col);
    };
    auto best_move = visited[state].back().first;

    best_moves.insert({current_move.state, best_move});

    reverse_move(best_move);

    while (state != starting_state) {
        bh = board_for_state(state);
        //print_board(bh);
        move(bh, best_move);
        state = state_for_board(bh);

        auto moves = find_moves(bh);
        auto current_depth = visited[state].back().second; // all entries should have the same depth.

        for (Move m : moves) {
            board_helper new_board = bh;
            move(new_board, m);
            auto new_state = state_for_board(new_board);
            // State is visited and depth is less.
            if (visited.count(new_state) && visited[new_state].back().second < current_depth) {
                // Found a solution path.
                best_move = m;
                reverse_move(m);
                best_moves.insert({new_state, m});

                break;
            }
        }
    }
    reverse_move(best_move);
    best_moves.insert({state, best_move});
}

GinkoSolver::GinkoSolver(board_state_t state) {
    best_moves.clear();
    solve(state);
}

solution_t GinkoSolver::get_solution() const {
    return best_moves;
}

