/* 
 * File:   TileBoard.cpp
 * Author: martin
 * 
 * Created on February 14, 2018, 10:37 AM
 */

#include "TileBoard.h"
#include <map>
#include <string>
#include <algorithm>
#include <tuple>

MotionDetector::Direction get_direction(Move m){
    if(m.from_row_col.first < m.to_row_col.first){
        return MotionDetector::DOWN;
    }
    if(m.from_row_col.first > m.to_row_col.first){
        return MotionDetector::UP;
    }
    if(m.from_row_col.second < m.to_row_col.second){
        return MotionDetector::RIGHT;
    }
    return MotionDetector::LEFT;    
}

TileBoard::TileBoard(Setup config) {
    setup_tiles();
}

void TileBoard::setup_tiles() {
    for(auto tile:board){
        remove(*tile);
        //delete tile;
    }
    board.clear();
    insert_tile(new Tile(VERTICAL), 0, 0);
    insert_tile(new Tile(BIG), 0, 1);
    insert_tile(new Tile(VERTICAL), 0, 3);

    insert_tile(new Tile(VERTICAL), 2, 0);
    insert_tile(new Tile(HORIZONTAL), 2, 1);
    insert_tile(new Tile(VERTICAL), 2, 3);

    insert_tile(new Tile(SQUARE), 4, 0);
    insert_tile(new Tile(SQUARE), 3, 1);
    insert_tile(new Tile(SQUARE), 3, 2);
    insert_tile(new Tile(SQUARE), 4, 3);    
    insert_tile(new Tile(EMPTY), 4, 1);
    insert_tile(new Tile(EMPTY), 4, 2);
    set_images(images);
    show_all();
}


void TileBoard::set_images(std::map<TileType, Glib::RefPtr<Gdk::Pixbuf> > imagemap) {
    images = imagemap;
    for(auto tile: board){
        tile->set(images[tile->get_tile_type()]);
    }
}


void TileBoard::move(Gtk::Widget* widget, MotionDetector::Direction d, double speed) {
    int row, col;
    auto tile = dynamic_cast<Tile*>(widget);
    if(!tile){
        return;
    }
    std::tie(row, col) = get_tile_row_col(tile);
    move(row, col, d, speed);
}


bool TileBoard::move(int row, int col, MotionDetector::Direction to, double speed) {
    auto tile = dynamic_cast<Tile*> (widget_at(col * col_grid_width(), row * row_grid_width()));
    if (!tile) return false;

    std::tie(row, col) = get_tile_row_col(tile);
    // max 2 points has to be checked.
    std::pair<int, int> target1;
    std::pair<int, int> target2;
    bool high = tile->get_row_span() > 1;
    bool wide = tile->get_col_span() > 1;
    
    switch (to) {
        case MotionDetector::UP:
            if (row <= 0) {
                return false;
            }
            target1 = {row - 1, col};
            target2 = target1;
            if (wide) {
                target2 = {row - 1, col + 1};
            }
            break;
        case MotionDetector::DOWN:
            if (row + high >= grid_rows - 1) { // bool is 0 or 1 here
                return false;
            }
            target1 = {row + 1 + high, col};
            target2 = target1;
            if (wide) {
                target2 = {row + 1 + high, col + 1};
            }
            break;
        case MotionDetector::LEFT:
            if (col == 0) {
                return false;
            }
            target1 = {row, col - 1};
            target2 = target1;
            if (high) {
                target2 = {row + 1, col - 1};
            }
            break;
        case MotionDetector::RIGHT:
            if (col + wide >= grid_columns - 1) {
                return false;
            }
            target1 = {row, col + 1 + wide};
            target2 = target1;
            if (high) {
                target2 = {row + 1, col + 1 + wide};
            }
    }
    if (is_occupied(target1.first, target1.second) || 
            is_occupied(target2.first, target2.second)){
        return false;
    }

    enable_mouse_move(false);
    FixedAnimationContainer::move(tile, to, speed);
    if(on_move)
        on_move();
    return true;
}

void TileBoard::on_animation_complete() {
    fix_empty_tiles();
    if(is_winning()){
        if(on_win)
            on_win();
    }
    enable_mouse_move(true);
}

void TileBoard::fix_empty_tiles() {
    std::vector<Tile*> empties;
    for(auto tile: board){
        if(tile->get_tile_type() == EMPTY){
            empties.push_back(tile);
        }
    }
    auto empty_it = empties.begin();
    for(int row = 0; row < grid_rows; row++){
        for(int col = 0; col < grid_columns; col++){
            if(!is_occupied(row, col)){
                put_in_grid(*(*empty_it++), row, col);
            }
        }
    }
}


void TileBoard::set_state(board_state_t state) {
    for (auto tile : board) {

    }
}

bool TileBoard::is_occupied(int look_at_row, int look_at_col) {
    for (int col = 0; col < grid_columns; col++) {
        for (int row = 0; row < grid_rows; row++) {
            if (auto tile = get_tile_at(row, col)) {
                if (tile &&
                        look_at_col >= col && look_at_col < col + tile->get_col_span() &&
                        look_at_row >= row && look_at_row < row + tile->get_row_span()) {
                    return tile->get_tile_type() != EMPTY;
                }
            }
        }
    }
    return false;
}

std::pair<char, char> TileBoard::get_tile_row_col(Tile* tile) {
    for (int col = 0; col < grid_columns; col++) {
        for (int row = 0; row < grid_rows; row++) {
            if (widget_at(col * col_grid_width(), row * row_grid_width()) == tile) {
                return std::pair<char, char>(row, col);
            }
        }
    }
    return {};
}

Tile* TileBoard::get_tile_at(int find_row, int find_col) {
    std::vector<Tile*> visited;
    for (int row = 0; row < this->grid_rows; row++) {
        for (int col = 0; col < this->grid_columns; col++) {
            auto grid_size_x = get_width() / grid_columns;
            auto grid_size_y = get_height() / grid_rows;
            auto tile = dynamic_cast<Tile*> (widget_at(col * grid_size_x, row * grid_size_y));
            if (std::find(visited.begin(), visited.end(), tile) == visited.end()) {
                if (find_row == row && find_col == col) {
                    return tile;
                }
                visited.push_back(tile);
            }
        }
    }
    return nullptr;
}

board_state_t TileBoard::state() {
    std::vector<Tile*> visited;
    std::string string_state;
    for (int row = 0; row < this->grid_rows; row++) {
        for (int col = 0; col < this->grid_columns; col++) {
            auto grid_size_x = get_width() / grid_columns;
            auto grid_size_y = get_height() / grid_rows;
            auto tile = dynamic_cast<Tile*> (widget_at(col * grid_size_x, row * grid_size_y));
            if (std::find(visited.begin(), visited.end(), tile) != visited.end())
                continue;
            visited.push_back(tile);
            string_state += state_for_tile(tile->get_tile_type());
        }
    }
    return string_state;
}

void TileBoard::insert_tile(Tile* tile, int row, int col) {
    board.push_back(tile);
    put_in_grid(*tile, row, col, tile->get_col_span(), tile->get_row_span());
}

bool TileBoard::is_winning() {
    if(get_tile_at(3, 1)){
        return get_tile_at(3, 1)->get_tile_type() == BIG;
    }
    return false;
}

solution_t TileBoard::get_solution() {
    solver.reset(new GinkoSolver(state()));
    return solver->get_solution();
}

bool TileBoard::is_interactive() const {
    return interactive;
}

void TileBoard::set_interactive(bool interactive) {
    FixedAnimationContainer::enable_mouse_move(interactive);
    this->interactive = interactive;
}
