/* 
 * File:   Tween.h
 * Author: Martin Kjellqvist <martin.kjellqvist@miun.se>
 *
 * Created on January 7, 2018, 7:28 AM
 */

#ifndef TWEEN_H
#define TWEEN_H

#include <sigc++-2.0/sigc++/functors/slot.h>
#include <glibmm.h>
#include <functional>
#include <cassert>

enum tween_type {
    LINEAR,
    DROP,
    BOUNCE
};
using tween = std::function<void()>;
using tween_complete_callback = std::function<void()>;

/**
 * time in ms per step.
 */
const int step_interval = 10;

template<typename PROP>
struct linear_tween {

    linear_tween(int range, unsigned int duration_ms, PROP prop, 
            tween_complete_callback complete_callback = tween_complete_callback())
    : range(range), prop(prop),
    start_value(prop.get_value()), complete_callback(complete_callback) {
        current = start_value;
        step = range * step_interval;
        step /= duration_ms;
    }

    bool operator()() {
        current += step;
        auto abs = [](int a) {
            return (a >= 0 ? a : -a);
        };
        if (abs(current - start_value) >= abs(range)) {
            prop.set_value(start_value + range);
            complete_callback();
            return false;
        }
        prop.set_value(current);
        return true;
    }
    const int range;
    PROP prop;
    const int start_value;
    tween_complete_callback complete_callback;
    double step;
    double current;    
};

template<typename PROP>
struct drop_tween {

    drop_tween(int range, unsigned int duration_ms, PROP prop, 
            tween_complete_callback complete_callback = tween_complete_callback())
    : range(range), prop(prop),
    start_value(prop.get_value()), complete_callback(complete_callback) {
        current = start_value;
        step = 0;
        auto steps = duration_ms / step_interval;
        if (steps == 0)
            steps = 1;        
        accel = range * 2.0 / (steps * steps);
    }

    bool operator()() {
        step += accel;
        current += step;
        auto abs = [](int a) {
            return (a >= 0 ? a : -a);
        };
        if (abs(current - start_value) >= abs(range)) {
            prop.set_value(start_value + range);
            complete_callback();
            return false;
        }
        prop.set_value(current);
        return true;
    }
    const int range;
    PROP prop;
    const int start_value;
    tween_complete_callback complete_callback;
    double accel;
    double step;
    double current;
};


/**
 * @tparam PROP type for property reflecting the value to animate.
 * @detail Usage: <tt>bounce_tween&lt;decltype(prop)&gt; tween(range, duration, prop);
 * prop is required to implement members get_value():numeric set_value(numeric) 
 */
template<typename PROP>
struct bounce_tween {
/**
 * 
 * @param range
 * @param duration_ms
 * @param prop
 */
    bounce_tween(int range, unsigned int duration_ms, PROP prop, 
            tween_complete_callback complete_callback = tween_complete_callback())
    : range(range), prop(prop),
    start_value(prop.get_value()), complete_callback(complete_callback) {
        current = start_value;
        step = 0.0;
        bounced = false;
        assert(duration_ms != 0);
        auto steps = duration_ms / step_interval;
        if (steps == 0)
            steps = 1;
        accel = range * 2.0 / (steps * steps);
    }

    bool operator()() {
        step += accel;
        current += step;
        auto abs = [](int a) {
            return (a >= 0 ? a : -a);
        };
        if (!bounced && abs(current - start_value) >= abs(range)) {
            accel = -4.0 * accel;
            bounced = true;
        }else
        // Passed range on the way back.
        if (bounced && abs(current - start_value) <= abs(range)) {
            prop.set_value(start_value + range);
            bounced = false;
            complete_callback();
            return false;
        }

        prop.set_value(current);
        return true;
    }
    const int range;
    PROP prop;
    const int start_value;
    tween_complete_callback complete_callback;
    
    double accel;
    double step;
    double current;
    bool bounced;
};

/**
 * tween_property is required to provide get_value():int and set_value(int).
 * @param range, negative values allowed
 * @param duration_ms
 * @param tt acceleration profile.
 * @param tween_property copyable.
 * @param callback, when the tween is complete, call this. void() function
 * @return function object to start the tween. lifetime needs to be preserved 
 * for the duration of the tween.
 */
template<typename PROP>
tween make_tween(int range, unsigned int duration_ms,
        PROP tween_property, tween_type tt = LINEAR, tween_complete_callback callback = tween_complete_callback()) {

    std::function<bool() > tt_tween;
    switch (tt) {
        case LINEAR:
            tt_tween = linear_tween<PROP>(range, duration_ms, tween_property, callback);
            break;
        case DROP:
            tt_tween = drop_tween<PROP>(range, duration_ms, tween_property, callback);
            break;
        case BOUNCE:
            tt_tween = bounce_tween<PROP>(range, duration_ms, tween_property, callback);
            break;
        default:
            assert(!"Invalid tween type");
    }
    tween tw = [tt_tween]() {
        auto tween_func = sigc::slot<bool()>(tt_tween);

        Glib::signal_timeout().connect(tween_func, step_interval, Glib::PRIORITY_DEFAULT_IDLE);
    };
    return tw;
}

#endif /* TWEEN_H */

