/*
 * File:   main.cpp
 * Author: Martin Kjellqvist <martin.kjellqvist@miun.se>
 *
 * Created on November 15, 2017, 10:50 AM
 */
#include <gtkmm.h>
#include <glibmm.h>

#include "TileBoard.h"
#include <iostream>
/*
 * The infamous tile game from the office.
 */
auto app = Gtk::Application::create("se.miun.markje.officetiles");
auto builder = Gtk::Builder::create_from_file("Ginko.glade");

Gtk::ApplicationWindow* window;
TileBoard tileboard;

Gtk::Toolbar* toolbar;
Gtk::ToolButton* playbutton = new Gtk::ToolButton;
Gtk::ToolButton* stopbutton = new Gtk::ToolButton;

Gtk::Label* statusmoveslabel;
Gtk::Label* statustimerlabel;

template<typename SIG_FUNC>
void put_in_toolbar(Gtk::Toolbar* bar, Gtk::ToolButton* toolButton, SIG_FUNC func, int position = -1) {
    auto callback = [func]() {
        func();
    };
    bar->insert(*toolButton, position, callback);
}

void display_status_moves(int moves){
    std::string text = "Moves: ";
    text += std::to_string(moves);
    statusmoveslabel->set_text(text);
}
struct GameState {

    GameState() : moves(0), time_elapsed(0), cheated(false), running(false) {
    }
    int moves;
    double time_elapsed;
    bool cheated;
    bool running;
} gamestate;

struct SolverState {
    bool running = false;
    solution_t current_solution;
    std::vector<board_state_t> shown_steps;
    size_t current_step = 0;
} solverstate;

void stop_solver_timer(){
    gamestate.running = false;
}

/**
 * 
 * @param data
 * @return false for the timer to stop
 */
bool next_step_in_solution(){
    if(!solverstate.running){
        //stop_solver_timer();
        return false;
    }
    if(tileboard.is_winning()){
        return false;
    }
    if(solverstate.current_step >= solverstate.shown_steps.size()){
        solverstate.shown_steps.push_back(tileboard.state());
    }
    board_helper bh = board_for_state(solverstate.shown_steps.back());
    Move m = solverstate.current_solution[solverstate.shown_steps[solverstate.current_step++]];
    if(!tileboard.move(m.from_row_col.first, m.from_row_col.second, get_direction(m), 3)){
        std::cerr << "Couldn't move from " << (int)m.from_row_col.first << ", " << (int)m.from_row_col.second << " to " << 
                (int)m.to_row_col.first << ", " << (int)m.to_row_col.second << "\n";
        return false;
    }
    display_status_moves(solverstate.current_step);
    return true;
}

void on_new_game() {
    gamestate.running = false;
    gamestate.moves = 0;
    gamestate.cheated = false;
    gamestate.time_elapsed = 0.0;
    tileboard.set_interactive(true);
    tileboard.setup_tiles();
}

void on_show_solver() {
    gamestate.cheated = true;
    gamestate.running = false;
    
    Gtk::HeaderBar* headerbar;
    builder->get_widget("solver-header", headerbar);
    window->set_titlebar(*headerbar);
    solverstate.current_solution = tileboard.get_solution();
    solverstate.current_step = 0;
    solverstate.shown_steps.clear();
    
    statusmoveslabel->set_text("Moves left: 0");
    std::string time_text = "States checked: ";
    time_text += std::to_string(solverstate.current_solution.size());
    statustimerlabel->set_text(time_text);
    tileboard.set_interactive(false);
}

void on_solver_first() {

}

void on_solver_play();
void on_solver_stop();

void on_solver_play() {
    toolbar->remove(*playbutton);
    *playbutton = Gtk::ToolButton();
    Gtk::Image* image;
    builder->get_widget("play-image", image);
    playbutton->set_icon_widget(*image);
    put_in_toolbar(toolbar, stopbutton, on_solver_stop, 2);
    toolbar->show_all();
    
    solverstate.running = true;
    auto mover = [](){ return next_step_in_solution(); };
    mover();
    Glib::signal_timeout().connect(mover, 400);
}

void on_solver_stop() {
    toolbar->remove(*stopbutton);
    *stopbutton = Gtk::ToolButton();
    Gtk::Image* image;
    builder->get_widget("stop-image", image);
    stopbutton->set_icon_widget(*image);
    put_in_toolbar(toolbar, playbutton, on_solver_play, 2);
    toolbar->show_all();
    
    solverstate.running = false;
    
}

void on_solver_back() {

}

void on_solver_last() {

}

void on_show_game_controls() {
    Gtk::HeaderBar* headerbar;
    builder->get_widget("game-header", headerbar);
    window->set_titlebar(*headerbar);
    tileboard.set_interactive(true);
    solverstate.running = false;
}
bool on_update_timer(){
    gamestate.time_elapsed += 0.1;
    int seconds = gamestate.time_elapsed;
    int tenths = gamestate.time_elapsed * 10 - 10 * seconds;
    int minutes = seconds / 60;
    seconds %= 60;
    std::string timestring = "Elapsed time: " + std::to_string(minutes) + 
            ":" + ((seconds<10)?"0":"") + std::to_string(seconds) + "." +
            std::to_string(tenths);
    statustimerlabel->set_text(timestring);
    return gamestate.running;
}
void on_piece_moved() {
    gamestate.moves++;
    std::string movestext = "Moves: " + std::to_string(gamestate.moves);
    
    statusmoveslabel->set_text(movestext);
    if(gamestate.running){
       return; 
    }
    gamestate.running = true;
    auto timer_updater = [](){  return on_update_timer();    };
    Glib::signal_timeout().connect(timer_updater, 100);

}

void on_victory() {
    gamestate.running = false;
    tileboard.set_interactive(false);
}

void connect_signals() {
    Gtk::Button * button;

    builder->get_widget("solve-button", button);
    auto show_solver_connector = []() {
        on_show_solver();
    };
    button->signal_clicked().connect(show_solver_connector, true);
}

template<typename SIG_FUNC>
void insert_to_toolbar(Gtk::Toolbar* bar, std::string image_name, SIG_FUNC func, int position = -1) {
    Gtk::Image* image;
    builder->get_widget(image_name, image);

    Gtk::ToolButton* toolButton = new Gtk::ToolButton;
    toolButton->set_icon_widget(*image);
    put_in_toolbar(bar, toolButton, func, position);
}

void setup_toolbar() {
    builder->get_widget("solver-toolbar", toolbar);

    insert_to_toolbar(toolbar, "first-image", on_solver_first);
    insert_to_toolbar(toolbar, "back-image", on_solver_back);
    // Need references to play and stop button
    Gtk::Image* image;
    builder->get_widget("stop-image", image);
    stopbutton->set_icon_widget(*image);
    builder->get_widget("play-image", image);
    playbutton->set_icon_widget(*image);

    put_in_toolbar(toolbar, playbutton, on_solver_play);

    insert_to_toolbar(toolbar, "last-image", on_solver_last);
    Gtk::Button * back_button;
    builder->get_widget("show-play-controls-button", back_button);
    auto show_play = []() {
        on_show_game_controls();
    };
    back_button->signal_clicked().connect(show_play, true);
    
    Gtk::Button* button; 
    builder->get_widget("new-button", button);
    auto on_new = [](){         on_new_game(); };
    button->signal_clicked().connect(on_new);
    
    toolbar->show_all();
}

void setup_application_window() {
    builder->get_widget("app-window", window);

    Gtk::HeaderBar* headerbar;
    builder->get_widget("game-header", headerbar);
    window->set_titlebar(*headerbar);

    builder->get_widget("moves-label", statusmoveslabel);
    builder->get_widget("timer-label", statustimerlabel);

    Gtk::EventBox* eventbox;
    builder->get_widget("frame-event-box", eventbox);

    std::map<TileType, Glib::RefPtr<Gdk::Pixbuf>> imagemap;

    Gtk::Image* image;
    builder->get_widget("square-image", image);
    imagemap.insert({SQUARE, image->get_pixbuf()});
    builder->get_widget("big-image", image);
    imagemap.insert({BIG, image->get_pixbuf()});
    builder->get_widget("vertical-image", image);
    imagemap.insert({VERTICAL, image->get_pixbuf()});
    builder->get_widget("horizontal-image", image);
    imagemap.insert({HORIZONTAL, image->get_pixbuf()});
    
    tileboard.set_images(imagemap);
    
    tileboard.show_all();

    eventbox->add(tileboard);
    eventbox->add_events(Gdk::POINTER_MOTION_MASK);

    tileboard.connect_detector(eventbox);
    tileboard.on_move = [](){   on_piece_moved(); };
    tileboard.on_win = [](){    on_victory(); };
    window->set_resizable(false);
    window->show_all();
}

int main() {
    setup_application_window();
    setup_toolbar();
    connect_signals();
    app->run(*window);
}

